/***************************************************************************
*             __________               __   ___.
*   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
*   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
*   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
*   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
*                     \/            \/     \/    \/            \/
* $Id: video.c,v 1.7 2004/01/15 07:48:02 hohensoh Exp $
*
* Plugin for LCD refresh test
*
* Copyright (C) 2004 Jrg Hohensohn aka [IDC]Dragon
*
* All files in this archive are subject to the GNU General Public License.
* See the file COPYING in the source tree root for full license agreement.
*
* This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
* KIND, either express or implied.
*
****************************************************************************/


/****************** imports ******************/

#include "plugin.h"
#include "sh7034.h"
#include "system.h"

#ifndef SIMULATOR /* not for simulator by now */
#ifdef HAVE_LCD_BITMAP /* and definitely not for the Player, haha */


/****************** constants ******************/

#define IMIA4 (*((volatile unsigned long*)0x09000180)) // timer 4

#define SCREENSIZE (LCD_WIDTH*LCD_HEIGHT/8) /* in bytes */


/****************** prototypes ******************/

void timer_set(unsigned period); // setup ISR and timer registers
void timer4_isr(void) __attribute__((interrupt_handler)); // IMIA4 ISR


/****************** globals ******************/

static struct plugin_api* rb; /* here is a global api struct pointer */

// buffer information
static struct
{
    int bufsize;
    unsigned char* pBufStart; // start of ring buffer
    unsigned char* pBufEnd; // end of ring buffer
    unsigned char* pBuf; // current position
    unsigned char* pOSD; // OSD memory (112 bytes for 112*8 pixels)
    int nTimeOSD; // OSD should stay for this many frames
    bool bDirtyOSD; // OSD needs redraw
} gBuf;


/****************** implementation ******************/

// setup ISR and timer registers
void timer_set(unsigned period)
{
    if (period)
    {
        and_b(~0x10, &TSTR); // Stop the timer 4
        and_b(~0x10, &TSNC); // No synchronization
        and_b(~0x10, &TMDR); // Operate normally

        IMIA4 = (unsigned long)timer4_isr; // install ISR

        TSR4 &= ~0x01;
        TIER4 = 0xF9; // Enable GRA match interrupt

        GRA4 = (unsigned short)(period - 1);
        TCR4 = 0x22; // clear at GRA match, sysclock/4
        IPRD = (IPRD & 0xFF0F) | 0x0010; // interrupt priority 1 (lowest)
        or_b(0x10, &TSTR); // start timer 4
    }
    else
    {
        and_b(~0x10, &TSTR); // stop the timer 4
        IPRD = (IPRD & 0xFF0F); // disable interrupt
    }
}


// timer interrupt handler to flip frames
void timer4_isr(void) // IMIA4
{
    int height; // height to display

    TSR4 &= ~0x01; // clear the interrupt
    
    height = LCD_HEIGHT/8;
    if (gBuf.nTimeOSD > 0)
    {
        gBuf.nTimeOSD--;
        height--;
    }

    // draw screen;  
    rb->lcd_blit(gBuf.pBuf, 0, 0, 
        LCD_WIDTH, height, LCD_WIDTH);

    if (gBuf.bDirtyOSD)
    {   // OSD to bottom line
        rb->lcd_blit(gBuf.pOSD, 0, LCD_HEIGHT/8-1, 
            LCD_WIDTH, 1, LCD_WIDTH);
        gBuf.bDirtyOSD = false;
    }

    gBuf.pBuf += SCREENSIZE;
    if (gBuf.pBuf >= gBuf.pBufEnd)
        gBuf.pBuf = gBuf.pBufStart;
}


int main(void)
{
    char buf[32];
    int button;
    unsigned period = FREQ/4 / 71; // default FPS
    unsigned oldperiod;
    unsigned acceleration = 2;

    rb->lcd_setfont(FONT_SYSFIXED);

    // init buffer
    rb->memset(&gBuf, 0, sizeof(gBuf));
    gBuf.pOSD = rb->lcd_framebuffer; // first screen line
    gBuf.pBufStart = rb->plugin_get_buffer(&gBuf.bufsize) + 100;
    gBuf.bufsize -= 100;
    if (gBuf.bufsize < 2*SCREENSIZE)
        return (PLUGIN_ERROR);
    gBuf.pBufEnd = gBuf.pBufStart + 2*SCREENSIZE;

    // draw frames
    gBuf.pBuf = gBuf.pBufStart;
    rb->memset(gBuf.pBuf, 0, SCREENSIZE); // white
    gBuf.pBuf += SCREENSIZE;
    rb->memset(gBuf.pBuf, 0xFF, SCREENSIZE); // black
    gBuf.pBuf = gBuf.pBufStart;

    timer_set(period); // start display interrupt
    do
    {
        rb->yield();
        button = rb->button_get(false);
        oldperiod = period;
        switch (button)
        {
        case BUTTON_UP:
            period--; // shorter period makes FPS faster
            acceleration = 5;
            break;

        case BUTTON_DOWN:
            period++; // longer period makes FPS slower
            acceleration = 5;
            break;

        case BUTTON_UP | BUTTON_REPEAT:
            period -= acceleration; // shorter period makes FPS faster
            acceleration += 5;
            if (acceleration > 300)
                acceleration = 300;
            break;

        case BUTTON_DOWN | BUTTON_REPEAT:
            period += acceleration; // longer period makes FPS slower
            acceleration += 5;
            if (acceleration > 300)
                acceleration = 300;
            break;
        }

        if (period < FREQ/4 / 99) // limit to 99 Hz
            period = FREQ/4 / 99;
        
        if (period > 65536) // limit to timer resolution
            period = 65536;
        
        if (oldperiod != period)
        {
            rb->snprintf(buf, sizeof(buf), "%d.%03d FPS (%d)",
                FREQ/4/period, 
                ((unsigned)FREQ*250/period)%1000,
                period);
            rb->lcd_puts(0, 0, buf);
            timer_set(period); // change the speed
            gBuf.nTimeOSD = 300;
            gBuf.bDirtyOSD = true;
        }
    } while (button != BUTTON_OFF && button != SYS_USB_CONNECTED);

    
    timer_set(0); // stop video ISR, now I can use the display again

    return (button == SYS_USB_CONNECTED) ? PLUGIN_USB_CONNECTED : PLUGIN_OK;
}


/***************** Plugin Entry Point *****************/

enum plugin_status plugin_start(struct plugin_api* api, void* parameter)
{
    int ret;
    (void)parameter;
    /* this macro should be called as the first thing you do in the plugin.
    it test that the api version and model the plugin was compiled for
    matches the machine it is running on */
    TEST_PLUGIN_API(api);
    
    rb = api; /* copy to global api pointer */
    
    /* now go ahead and have fun! */
    ret = main();
    if (ret==PLUGIN_USB_CONNECTED)
        rb->usb_screen();
    return ret;
}

#endif /* #ifdef HAVE_LCD_BITMAP */
#endif /* #ifndef SIMULATOR */

